# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import shutil

from flask import Blueprint
from flask import current_app
from flask import jsonify
from flask import request
from flask_login import current_user
from flask_login import logout_user
from sqlalchemy import and_
from sqlalchemy import or_
from sqlalchemy import select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.exc import SQLAlchemyError

from asreview.webapp import DB
from asreview.webapp._authentication.decorators import login_remote_user
from asreview.webapp._authentication.decorators import login_required
from asreview.webapp._authentication.models import Project
from asreview.webapp._authentication.models import User
from asreview.webapp._authentication.oauth_handler import OAuthHandler
from asreview.webapp._authentication.utils import has_email_configuration
from asreview.webapp._authentication.utils import perform_login_user
from asreview.webapp._authentication.utils import send_confirm_account_email
from asreview.webapp._authentication.utils import send_forgot_password_email

bp = Blueprint("auth", __name__, url_prefix="/auth")


def _signed_in_payload(user):
    return {
        "logged_in": True,
        "name": user.get_name(),
        "id": user.id,
        "role": user.role,
        "message": f"User {user.identifier} is logged in.",
    }


def _select_user_by_email(email):
    stmt = select(User).where(or_(User.identifier == email, User.email == email))
    return DB.session.execute(stmt).scalar_one_or_none()


# ------------------
#      ROUTES
# ------------------


@bp.route("/signin", methods=["POST"])
def signin():
    email = request.form.get("email").strip()
    password = request.form.get("password", "")

    # get the user
    user = _select_user_by_email(email)

    if not user:
        # user does not exsist
        result = (404, {"message": f"User account {email} does not exist."})
    elif not user.confirmed:
        # account is not confirmed
        result = (404, {"message": f"User account {email} is not confirmed."})
    else:
        # user exists and is confirmed: verify password
        if user.verify_password(password):
            if perform_login_user(user, current_app):
                result = (200, _signed_in_payload(user))
            else:
                result = (
                    401,
                    {"message": "Unable to login user with verified password."},
                )
        else:
            # password is wrong
            if user.origin == "asreview":
                # if this is an asreview user
                result = (404, {"message": f"Incorrect password for user {email}."})
            else:
                # this must be an OAuth user trying to get in with
                # a password
                service = user.origin.capitalize()
                result = (404, {"message": f"Please login with the {service} service."})

    status, message = result
    response = jsonify(message)
    return response, status


@bp.route("/signup", methods=["POST"])
def signup():
    user_id = None
    # Can we create accounts?
    if current_app.config.get("ALLOW_ACCOUNT_CREATION", False):
        email = request.form.get("email", "").strip()
        name = request.form.get("name", "").strip()
        affiliation = request.form.get("affiliation", "").strip()
        password = request.form.get("password")

        # Handle terms of agreement
        terms_config = current_app.config.get("TERMS_OF_AGREEMENT", False)
        if terms_config:
            # Feature enabled: validate checkbox was checked
            terms_accepted = (
                request.form.get("terms_accepted", "false").lower() == "true"
            )
            if not terms_accepted:
                return jsonify(
                    {
                        "message": "You must accept the terms of agreement to create an account."
                    }
                ), 400
        else:
            # Feature disabled: default to True (no terms to accept)
            terms_accepted = True

        # check if email already exists
        user = _select_user_by_email(email)

        # return error if user doesn't exist
        if isinstance(user, User):
            result = (403, f'User with email "{email}" already exists.')
        else:
            try:
                identifier = email
                origin = "asreview"
                # are we going to verify the email?
                email_verification = bool(
                    current_app.config.get("EMAIL_VERIFICATION", False)
                )
                # set confirmed to False if we 'do' verification. Note
                # that this route only creates 'asreview' accounts
                confirmed = not email_verification
                # create the User account
                user = User(
                    identifier=identifier,
                    origin=origin,
                    email=email,
                    name=name,
                    affiliation=affiliation,
                    password=password,
                    confirmed=confirmed,
                    terms_accepted=terms_accepted,
                )
                # if this is an un-confirmed account, set token
                if not confirmed:
                    # set token data
                    user = user.set_token()
                # store user
                DB.session.add(user)
                DB.session.commit()
                # set user id
                user_id = user.id
                # at this stage, if all went well, the User account is
                # stored in the database, send the verification email
                # if applicable
                if email_verification:
                    # send email
                    send_confirm_account_email(user, current_app, "create")
                    message = (
                        f"An email has been sent to {user.email} to verify "
                        "your account. Please follow instructions."
                    )
                    # result
                    result = (201, {"message": message, "user_id": user_id})
                else:
                    # immediately log user in
                    if perform_login_user(user, current_app):
                        result = (201, _signed_in_payload(user))
                    else:
                        result = (401, "Account confirmed, but unable to sign in.")
            except ValueError as e:
                result = (400, f"Unable to create your account! Reason: {str(e)}")
            except IntegrityError as e:
                DB.session.rollback()
                result = (403, f"Unable to create your account! Reason: {str(e)}")
            except SQLAlchemyError as e:
                DB.session.rollback()
                result = (403, f"Unable to create your account! Reason: {str(e)}")
    else:
        result = (404, "The app is not configured to create accounts")

    status, payload = result
    payload = {"message": payload} if isinstance(payload, str) else payload

    response = jsonify(payload)
    return response, status


@bp.route("/confirm_account", methods=["POST"])
def confirm_account():
    """Confirms account with email verification"""
    # https://realpython.com/handling-email-confirmation-in-flask/#handle-email-confirmation

    if not current_app.config.get("EMAIL_VERIFICATION", False):
        return "Email verification is not enabled", 400

    # find user by token and user id
    user_id = request.form.get("user_id", 0)
    token = request.form.get("token", "")

    user = User.query.filter(
        and_(User.id == user_id, User.token == token)
    ).one_or_none()

    if not user:
        result = (404, "No user account / correct token found.")
    elif not user.token_valid(token):
        message = (
            "Can not confirm account, token has expired. "
            + 'Use "forgot password" to obtain a new one.'
        )
        result = (403, message)
    else:
        user = user.confirm_user()
        try:
            # commit changes
            DB.session.commit()
            # sign user in
            if perform_login_user(user, current_app):
                result = (200, _signed_in_payload(user))
            else:
                result = (401, "Account confirmed, but unable to sign in.")
        except SQLAlchemyError:
            DB.session.rollback()
            result = (500, "Account not confirmed")

    status, payload = result
    payload = {"message": payload} if isinstance(payload, str) else payload

    response = jsonify(payload)
    return response, status


@bp.route("/get_profile", methods=["GET"])
@login_required
def get_profile():
    user = User.query.filter(User.id == current_user.id).one_or_none()
    if user:
        result = (
            200,
            {
                "identifier": user.identifier,
                "email": user.email,
                "origin": user.origin,
                "name": user.name,
                "affiliation": user.affiliation,
            },
        )
    else:
        result = (404, "No user found.")

    status, message = result
    response = jsonify({"message": message})
    return response, status


@bp.route("/forgot_password", methods=["POST"])
def forgot_password():
    user_id = None
    if has_email_configuration(current_app):
        # get email address from request
        email_address = request.form.get("email", "").strip()

        # check if email already exists
        user = _select_user_by_email(email_address)

        if not user:
            result = (404, f'User with email "{email_address}" not found.')
        elif user.origin != "asreview":
            result = (404, f"Your account has been created with {user.origin}.")
        else:
            # set a token
            user = user.set_token()
            try:
                # store data
                DB.session.commit()
                # send email
                send_forgot_password_email(user, current_app)
                # result
                result = (200, f"An email has been sent to {email_address}")
                # user_id is safe to set
                user_id = user.id

            except SQLAlchemyError as e:
                DB.session.rollback()
                result = (403, f"Unable to to confirm user! Reason: {str(e)}")
    else:
        result = (404, "Forgot-password feature is not used in this app.")

    status, message = result
    response = jsonify({"message": message, "user_id": user_id})
    return response, status


@bp.route("/reset_password", methods=["POST"])
def reset_password():
    """Resests password of user"""
    if has_email_configuration(current_app):
        new_password = request.form.get("password", "").strip()
        token = request.form.get("token", "").strip()
        user_id = request.form.get("user_id", "0").strip()
        user = User.query.filter(User.id == user_id).one_or_none()

        if not user:
            result = (
                404,
                "User not found, try restarting the forgot-password procedure.",
            )
        elif not user.token_valid(token):
            result = (
                404,
                "Token is invalid or too old, restart the forgot-password procedure.",
            )
        else:
            try:
                user = user.reset_password(new_password)
                DB.session.commit()
                # sign user in after password is reset
                if perform_login_user(user, current_app):
                    result = (200, _signed_in_payload(user))
                else:
                    result = (401, "Password reset, but unable to sign in.")
            except ValueError as e:
                DB.session.rollback()
                result = (500, f"Unable to reset your password! Reason: {str(e)}")
            except SQLAlchemyError as e:
                DB.session.rollback()
                result = (500, f"Unable to reset your password! Reason: {str(e)}")
    else:
        result = (404, "Reset-password feature is not used in this app.")

    status, payload = result
    payload = {"message": payload} if isinstance(payload, str) else payload

    response = jsonify(payload)
    return response, status


@bp.route("/update_profile", methods=["POST"])
@login_required
def update_profile():
    """Update user profile"""
    user = User.query.filter(User.id == current_user.id).one_or_none()
    old_email = user.email

    if user:
        email = request.form.get("email", None)
        name = request.form.get("name", None)
        affiliation = request.form.get("affiliation", None)
        old_password = request.form.get("old_password", None)
        new_password = request.form.get("new_password", None)
        reconfirm = (
            email != old_email
            and user.origin == "asreview"
            and current_app.config.get("EMAIL_VERIFICATION", False)
        )

        try:
            user = user.update_profile(
                email, name, affiliation, old_password, new_password, reconfirm
            )
            DB.session.commit()
            if reconfirm:
                # send email
                send_confirm_account_email(user, current_app, "change_email")
                # email has been changed and we verify email
                message = (
                    "User profile updated, but new email address needs verification."
                )
                result = (
                    200,
                    {"message": message, "email_changed": True, "user_id": user.id},
                )
            else:
                result = (200, "User profile updated.")
        except ValueError as e:
            DB.session.rollback()
            result = (400, f"Unable to update your profile! Reason: {str(e)}")
        except IntegrityError as e:
            DB.session.rollback()
            result = (500, f"Unable to update your profile! Reason: {str(e)}")
        except SQLAlchemyError as e:
            DB.session.rollback()
            result = (500, f"Unable to update your profile! Reason: {str(e)}")

    else:
        result = (404, "No user found")

    status, message = result
    if isinstance(message, str):
        message = {"message": message}
    response = jsonify(message)
    return response, status


@bp.route("/user", methods=["GET"])
@login_remote_user
@login_required
def user():
    return jsonify(
        {
            "name": current_user.get_name(),
            "id": current_user.id,
            "role": current_user.role,
        }
    ), 200


@bp.route("/signout", methods=["DELETE"])
@login_required
def signout():
    if current_user:
        identifier = current_user.identifier
        logout_user()
        result = (200, f"User with identifier {identifier} has been signed out.")
    else:
        result = (404, "No user found, no one can be signed out.")

    status, message = result
    response = jsonify({"message": message})
    return response, status


@bp.route("/oauth_callback", methods=["POST"])
def oauth_callback():
    # get parameters
    code = request.form.get("code", "").strip()
    provider = request.form.get("provider", "").strip()
    redirect_uri = request.form.get("redirect_uri", "").strip()

    # assuming we have this provider
    oauth_handler = current_app.config.get("OAUTH", False)
    if (
        isinstance(oauth_handler, OAuthHandler)
        and provider in oauth_handler.providers()
    ):
        # get user credentials for this user
        (identifier, email, name) = oauth_handler.get_user_credentials(
            provider, code, redirect_uri
        )
        # make sure identifier is a string
        identifier = str(identifier)
        # try to find this user
        user = User.query.filter(User.identifier == identifier).one_or_none()
        # flag for response (I'd like to communicate if this user was created)
        created_account = False
        # if not create user
        if user is None:
            # Check if terms of agreement are required
            terms_required = current_app.config.get("TERMS_OF_AGREEMENT", False)

            if terms_required:
                terms_accepted = request.form.get("terms_accepted", "false") == "true"
                if not terms_accepted:
                    message = "Terms of agreement must be accepted to create an account"
                    return jsonify({"message": message}), 400
            else:
                terms_accepted = True

            try:
                origin = provider
                confirmed = True
                user = User(
                    identifier=identifier,
                    origin=origin,
                    email=email,
                    name=name,
                    confirmed=confirmed,
                    terms_accepted=terms_accepted,
                )
                DB.session.add(user)
                DB.session.commit()
                created_account = True
            except IntegrityError:
                DB.session.rollback()
                message = (
                    "OAuth: integrity error, verify if you "
                    + "already have created an account!"
                )
                # return this immediately
                return jsonify({"message": message}), 409
            except SQLAlchemyError:
                DB.session.rollback()
                message = "OAuth: unable to create your account!"
                # return this immediately
                return jsonify({"message": message}), 409

        # log in the existing/created user immediately
        logged_in = perform_login_user(user, current_app)
        result = (
            200,
            {
                "account_created": created_account,
                "logged_in": logged_in,
                "name": user.get_name(),
                "id": user.id,
            },
        )
    else:
        result = (401, {"message": f"OAuth provider {provider} could not be found"})

    status, message = result
    response = jsonify(message)
    return response, status


@bp.route("/delete_account", methods=["DELETE"])
@login_required
def delete_account():
    """Delete the current logged-in user's account"""
    try:
        user = current_user

        # Block deletion if user is an admin and the only admin
        if user.is_admin:
            admin_count = DB.session.query(User).filter_by(role="admin").count()
            if admin_count <= 1:
                return (
                    jsonify(
                        {
                            "message": "Cannot delete the only admin account. "
                            "Please create another admin account first."
                        }
                    ),
                    400,
                )

        # Block deletion if user owns projects
        if Project.query.filter_by(owner_id=user.id).count() > 0:
            return (
                jsonify(
                    {
                        "message": "You still own projects. "
                        "Please let an admin transfer ownership or delete them before deleting your account."
                    }
                ),
                400,
            )

        # delete projects from ASReview folder
        user_projects = Project.query.filter_by(owner_id=user.id).all()
        for project in user_projects:
            if project.project_path.exists():
                shutil.rmtree(project.project_path)

        # delete user plus records in projects table
        DB.session.delete(user)
        DB.session.commit()

        # Logout the user after account deletion
        logout_user()

        return jsonify({"message": "Account deleted successfully"}), 200

    except SQLAlchemyError as e:
        DB.session.rollback()
        return jsonify({"message": f"Database error: {str(e)}"}), 500
    except Exception as e:
        DB.session.rollback()
        logging.error(f"Error deleting account: {e}")
        return jsonify({"message": f"Error deleting account: {str(e)}"}), 500



import pytest

from asreview.extensions import extensions
from asreview.extensions import load_extension


def test_classifiers():
    assert len(extensions("models.classifiers")) >= 1


@pytest.mark.parametrize("classifier", extensions("models.classifiers"))
def test_classifier_name(classifier):
    model = load_extension("models.classifiers", classifier.name)()
    assert model.name == classifier.name


@pytest.mark.parametrize("classifier", extensions("models.classifiers"))
def test_classifier_param(classifier):
    model = load_extension("models.classifiers", classifier.name)()
    assert isinstance(model.get_params(), dict)



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.models.models import AI_MODEL_CONFIGURATIONS
from asreview.models.models import get_ai_config

__all__ = [
    "balancers",
    "classifiers",
    "feature_extractors",
    "queriers",
    "stoppers",
    "AI_MODEL_CONFIGURATIONS",
    "get_ai_config",
]

"""Active learning model components.

Components like classifiers, query strategies, balance strategies, and
feature_extractor techniques."""



import pytest

import asreview.webapp._entry_points.lab as lab


@pytest.fixture
def mock_env_lab_testing(monkeypatch):
    monkeypatch.setenv("ASREVIEW_LAB_TESTING", "true")


def test_authentication_parameter(mock_env_lab_testing):
    app = lab.lab_entry_point(["--enable-auth"])
    assert app.config.get("AUTHENTICATION")


def test_without_authentication_parameter(mock_env_lab_testing):
    app = lab.lab_entry_point([])
    assert app.config.get("AUTHENTICATION") is False



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from asreview.learner import ActiveLearningCycleData

AI_MODEL_CONFIGURATIONS = [
    {
        # ELAS u4 is the default model for this version ASReview LAB. The model
        # parameters have been optimized on the SYNERGY dataset. After expert
        # elicitation, the model and parameters have been chosen.
        # Optimization setup: https://github.com/asreview/asreview-optuna/releases/tag/ASReview2_0b4-nb-tfidf-full-1
        "name": "elas_u4",
        "label": "ELAS u4",
        "type": "ultra",
        "value": ActiveLearningCycleData(
            querier="max",
            classifier="svm",
            classifier_param={"loss": "squared_hinge", "C": 0.11},
            balancer="balanced",
            balancer_param={"ratio": 9.8},
            feature_extractor="tfidf",
            feature_extractor_param={
                "ngram_range": (1, 2),
                "sublinear_tf": True,
                "min_df": 1,
                "max_df": 0.95,
            },
        ),
    },
    {
        # ELAS u3 mimics the performance of ASReview LAB version 1. The model
        # has an updated and optimized balancer, as "double" balance has been
        # deprecated in this version of ASReview LAB.
        "name": "elas_u3",
        "label": "ELAS u3",
        "type": "ultra",
        "value": ActiveLearningCycleData(
            querier="max",
            classifier="nb",
            classifier_param={"alpha": 3.822},
            balancer="balanced",
            balancer_param={"ratio": 1.2},
            feature_extractor="tfidf",
            feature_extractor_param={"stop_words": "english"},
        ),
    },
    # ELAS u2 refers to the model that was used in the ASReview LAB version 1. The
    # model is not available in the current version of ASReview LAB.
    # ELAS u1 refers to the model that was used in the ASReview LAB version 0. The
    # model is not available in the current version of ASReview LAB.
    {
        # ELAS l2 is the multilingual model for this version ASReview LAB. The model
        # parameters have been optimized on the SYNERGY dataset. After expert
        # elicitation, the model and parameters have been chosen.
        # Optimization setup: https://github.com/asreview/asreview-optuna/releases/tag/ASReview2_0b4-svm-e5-full-1
        "name": "elas_l2",
        "label": "ELAS l2",
        "type": "lang",
        "extensions": ["asreview-dory"],
        "value": ActiveLearningCycleData(
            querier="max",
            classifier="svm",
            classifier_param={"loss": "squared_hinge", "C": 0.106, "max_iter": 5000},
            balancer="balanced",
            balancer_param={"ratio": 9.707},
            feature_extractor="multilingual-e5-large",
            feature_extractor_param={"normalize": True},
        ),
    },
    {
        # ELAS h3 is the heavy model for this version ASReview LAB. The model
        # parameters have been optimized on the SYNERGY dataset. After expert
        # elicitation, the model and parameters have been chosen.
        # Optimization setup: https://github.com/asreview/asreview-optuna/releases/tag/ASReview2_0b4-svm-mxbai-full-1
        "name": "elas_h3",
        "label": "ELAS h3",
        "type": "heavy",
        "extensions": ["asreview-dory"],
        "value": ActiveLearningCycleData(
            querier="max",
            classifier="svm",
            classifier_param={"loss": "squared_hinge", "C": 0.067, "max_iter": 5000},
            balancer="balanced",
            balancer_param={"ratio": 9.724},
            feature_extractor="mxbai",
            feature_extractor_param={"normalize": True},
        ),
    },
]


def get_ai_config(name=None):
    """Get the AI configuration.

    Parameters
    ----------
    name: str
        The name of the AI configuration. If None, the default AI
        configuration is returned.

    Returns
    -------
    dict:
        The default AI configuration.
    """

    if name is None:
        return filter(
            lambda x: x["type"] == "ultra", AI_MODEL_CONFIGURATIONS
        ).__next__()

    for model_config in AI_MODEL_CONFIGURATIONS:
        if model_config["name"] == name:
            return model_config

    raise ValueError(f"Model configuration {name} not found.")



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Optional

from asreview.extensions import load_extension
from asreview.utils import _read_config_file

__all__ = []


def _unpack_params(params):
    return {} if params is None else params


def _get_name_from_estimator(estimator):
    """Get the name of the estimator.

    Parameters
    ----------
    estimator: object
        The estimator to get the name from.

    Returns
    -------
    str
        The name of the estimator.

    """
    if estimator is None:
        return None

    return estimator.name


@dataclass
class ActiveLearningCycleData:
    querier: str
    classifier: Optional[str] = None
    balancer: Optional[str] = None
    feature_extractor: Optional[str] = None
    stopper: Optional[str] = None
    querier_param: Optional[dict[str, Any]] = field(default_factory=dict)
    classifier_param: Optional[dict[str, Any]] = field(default_factory=dict)
    balancer_param: Optional[dict[str, Any]] = field(default_factory=dict)
    feature_extractor_param: Optional[dict[str, Any]] = field(default_factory=dict)
    stopper_param: Optional[dict[str, Any]] = field(default_factory=dict)
    n_query: int = 1


class ActiveLearningCycle:
    """Active learner cycle class.

    The active learner cycle class is a wrapper around the various learner components.

    The classifier is optional, if no classifier is provided, the active learner will
    only rank the instances based on the query strategy. This can be useful for example
    if you want random screening.

    Parameters
    ----------
    querier: BaseQueryStrategy
        The query strategy to use.
    classifier: BaseTrainClassifier
        The classifier to use. Default is None.
    balancer: BaseTrainClassifier
        The balance strategy to use. Default is None.
    feature_extractor: BaseFeatureExtraction
        The feature extraction method to use. Default is None.
    stopper: BaseStopper
        The stopping criteria. Default is None.
    n_query: int, callable
        The number of instances to query at once. Default is 1.

    """

    def __init__(
        self,
        querier,
        classifier=None,
        balancer=None,
        feature_extractor=None,
        stopper=None,
        n_query=1,
    ):
        self.querier = querier
        self.classifier = classifier
        self.balancer = balancer
        self.feature_extractor = feature_extractor
        self.stopper = stopper
        self.n_query = n_query

    def get_n_query(self, results, labels):
        """Get the number of records to query at each step in the active learning.

        n_query can be an integer or a function that takes the results of the
        simulation as input. If n_query is a function, it should return an integer.
        n_query can not be larger than the number of records left to label.

        Parameters
        ----------
        n_query: int | callable
            Number of records to query at each step in the active learning
            process. Default is 1.
        results: pd.DataFrame
            The results of the simulation.

        Returns
        -------
        int
            Number of records to query at each step in the active learning process.

        """

        n_query = self.n_query(results) if callable(self.n_query) else self.n_query
        return min(n_query, len(labels) - len(results))

    def transform(self, X):
        """Transform the data.

        Parameters
        ----------
        X: np.array
            The instances to transform.

        Returns
        -------
        np.array, scipy.sparse.csr_matrix:
            The transformed instances.
        """
        return self.feature_extractor.fit_transform(X)

    def fit(self, X, y):
        """Fit the classifier to the data.

        Parameters
        ----------
        X: np.array
            The instances to fit.
        y: np.array
            The labels of the instances.
        """
        if self.balancer is None:
            sample_weight = None
        else:
            sample_weight = self.balancer.compute_sample_weight(y)

        self.classifier.fit(X, y, sample_weight=sample_weight)
        return self

    def rank(self, X):
        """Rank the instances in X.

        Parameters
        ----------
        X: np.array
            The instances to rank.

        Returns
        -------
        np.array:
            The ranking of the instances.
        """

        if self.classifier is None:
            return self.querier.query(X)

        try:
            proba = self.classifier.predict_proba(X)
            return self.querier.query(proba[:, 1])
        except AttributeError:
            try:
                scores = self.classifier.decision_function(X)

                if "proba" in self.querier.get_params(deep=False):
                    self.querier.set_params(proba=False)

                return self.querier.query(scores)

            except AttributeError:
                raise AttributeError(
                    "Not possible to compute probabilities or "
                    "decision function for this classifier."
                )

    def stop(self, results, data):
        """Check if the stopping criteria is met.

        Parameters
        ----------
        results: pd.DataFrame
            The results of the simulation.
        data: pandas.DataFrame
            The data store object.

        Returns
        -------
        bool:
            True if the stopping criteria is met, False otherwise.
        """
        if self.stopper is None:
            return False
        return self.stopper.stop(results, data)

    @classmethod
    def from_meta(cls, cycle_meta_data, skip_feature_extraction=False):
        """Load the active learner from a metadata object.

        Parameters
        ----------
        cycle_meta_data: CycleMetaData
            The metadata object with the active learner settings.

        Returns
        -------
        ActiveLearningCycle:
            The active learner cycle object.
        """
        query_class = load_extension("models.queriers", cycle_meta_data.querier)
        query_model = query_class(**cycle_meta_data.querier_param)

        if cycle_meta_data.classifier is not None:
            classifier_class = load_extension(
                "models.classifiers", cycle_meta_data.classifier
            )
            classifier_model = classifier_class(**cycle_meta_data.classifier_param)
        else:
            classifier_model = None

        if cycle_meta_data.balancer is not None:
            balance_class = load_extension("models.balancers", cycle_meta_data.balancer)
            balance_model = balance_class(**cycle_meta_data.balancer_param)
        else:
            balance_model = None

        if (
            not skip_feature_extraction
            and cycle_meta_data.feature_extractor is not None
        ):
            feature_class = load_extension(
                "models.feature_extractors", cycle_meta_data.feature_extractor
            )
            feature_model = feature_class(**cycle_meta_data.feature_extractor_param)
        else:
            feature_model = None

        if cycle_meta_data.stopper is not None:
            stopper_class = load_extension("models.stoppers", cycle_meta_data.stopper)
            stopper_model = stopper_class(**cycle_meta_data.stopper_param)
        else:
            stopper_model = None

        return cls(
            querier=query_model,
            classifier=classifier_model,
            balancer=balance_model,
            feature_extractor=feature_model,
            stopper=stopper_model,
            n_query=cycle_meta_data.n_query,
        )

    @classmethod
    def from_file(cls, fp, load=None):
        """Load the active learner from a file.

        Parameters
        ----------
        fp: str, Path
            Review config file.
        load: object
            Config reader. Default tomllib.load for TOML (.toml) files,
            otherwise json.load.
        """
        if load is not None:
            with open(fp, "r") as f:
                return cls.from_meta(ActiveLearningCycleData(**load(f)))

        return cls.from_meta(ActiveLearningCycleData(**_read_config_file(fp)))

    def to_meta(self):
        return ActiveLearningCycleData(
            querier=_get_name_from_estimator(self.querier),
            querier_param=self.querier.get_params(deep=False),
            classifier=_get_name_from_estimator(self.classifier),
            classifier_param=self.classifier.get_params(deep=False)
            if self.classifier is not None
            else None,
            balancer=_get_name_from_estimator(self.balancer),
            balancer_param=self.balancer.get_params(deep=False)
            if self.balancer is not None
            else None,
            feature_extractor=_get_name_from_estimator(self.feature_extractor),
            feature_extractor_param=self.feature_extractor.get_params(deep=False)
            if self.feature_extractor is not None
            else None,
            stopper=_get_name_from_estimator(self.stopper),
            stopper_param=self.stopper.get_params(deep=False)
            if self.stopper is not None
            else None,
            n_query=self.n_query,
        )

    def to_file(self, fp):
        with open(fp, "w") as f:
            json.dump(asdict(self.to_meta()), f)



import json
import logging
import multiprocessing as mp
import socket
import threading
import signal
import errno
from collections import deque

from sqlalchemy import create_engine
from sqlalchemy import text
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker

from asreview.webapp import asreview_path
from asreview.webapp._task_manager.models import Base
from asreview.webapp._task_manager.models import ProjectQueueModel
from asreview.webapp._tasks import run_task

DEFAULT_TASK_MANAGER_HOST = "localhost"
DEFAULT_TASK_MANAGER_PORT = 5101
DEFAULT_TASK_MANAGER_WORKERS = 2

# Set up a module-level logger for the task manager
logger = logging.getLogger("asreview.task_manager")


def _setup_logging(verbose=0):
    """Set up logging for the task manager."""

    # Determine log level from verbosity count
    if verbose >= 2:
        log_level = "DEBUG"
    elif verbose == 1:
        log_level = "INFO"
    else:
        log_level = None

    if log_level is not None:
        numeric_level = getattr(logging, log_level.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError(f"Invalid log level: {log_level}")

        logger.setLevel(numeric_level)

    logger.propagate = False
    logger.handlers.clear()

    handler = logging.StreamHandler()
    formatter = logging.Formatter("%(levelname)s:%(name)s:%(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)


class RunModelProcess(mp.Process):
    def __init__(
        self,
        func,
        args=(),
        host=DEFAULT_TASK_MANAGER_HOST,
        port=DEFAULT_TASK_MANAGER_PORT,
    ):
        super().__init__()
        self.func = func
        self.args = args
        self.host = host
        self.port = port

    def run(self):
        _setup_logging(verbose=1)
        payload = {"action": "remove", "project_id": self.args[0]}
        try:
            self.func(*self.args)
        except Exception:
            logger.exception(f"Exception in task for project {self.args[0]}")
            payload["action"] = "failure"
        finally:
            if self.host and self.port:
                self._send_payload(payload)

    def _send_payload(self, payload):
        """Send a payload to the task manager via socket."""
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as client_socket:
            try:
                client_socket.connect((self.host, self.port))
                client_socket.sendall(json.dumps(payload).encode("utf-8"))
            except Exception as e:
                logger.error(f"Failed to send payload: {e}")


class TaskManager:
    def __init__(
        self,
        max_workers=DEFAULT_TASK_MANAGER_WORKERS,
        host=DEFAULT_TASK_MANAGER_HOST,
        port=DEFAULT_TASK_MANAGER_PORT,
    ):
        self.running_processes = {}  # project_id -> Process object
        self.max_workers = int(max_workers)

        # set up parameters for socket endpoint
        self.host = host
        self.port = int(port)
        self.message_buffer = deque()
        self.receive_bytes = 1024  # bytes read when receiving messages
        self.timeout = 0.1  # wait for 0.1 seconds for incoming messages

        # set up database
        database_url = f"sqlite:///{asreview_path()}/queue.sqlite"
        engine = create_engine(database_url, pool_size=5, max_overflow=10)
        Base.metadata.create_all(engine)

        Session = sessionmaker(bind=engine)
        self.session = Session()

        # Handle migration for created_at column
        self._migrate_created_at_column()

        self.client_thread = None
        self.client_conn = None

    def _migrate_created_at_column(self):
        """Add created_at column if it doesn't exist (for existing deployments)."""
        try:
            # Check if created_at column exists
            result = self.session.execute(text("PRAGMA table_info(queue)"))
            columns = [row[1] for row in result.fetchall()]

            if "created_at" not in columns:
                logger.info("Adding created_at column to queue table")
                # Add the column with a default value for existing records
                self.session.execute(
                    text(
                        "ALTER TABLE queue ADD COLUMN created_at DATETIME DEFAULT CURRENT_TIMESTAMP"
                    )
                )
                self.session.commit()
                logger.info("created_at column added successfully")
            else:
                logger.debug("created_at column already exists")
        except Exception as e:
            logger.error(f"Failed to migrate created_at column: {e}")
            self.session.rollback()

    @property
    def waiting(self):
        records = self.session.query(ProjectQueueModel).all()
        return [r.project_id for r in records]

    def insert_in_waiting(self, project_id, simulation):
        try:
            new_record = ProjectQueueModel(
                project_id=str(project_id), simulation=bool(simulation)
            )
            self.session.execute(text("BEGIN TRANSACTION"))
            self.session.add(new_record)
            self.session.commit()
        except IntegrityError:
            logger.info(f"Task for project {project_id} already in queue")
            self.session.rollback()
        except Exception:
            logger.error(f"Failed to add task for project {project_id} queue")
            self.session.rollback()
        else:
            logger.info(f"Task for project {project_id} inserted into queue")

    def is_waiting(self, project_id):
        return (
            self.session.query(ProjectQueueModel)
            .filter_by(project_id=project_id)
            .first()
            or False
        )

    def is_pending(self, project_id):
        return project_id in self.running_processes

    def remove_pending(self, project_id):
        if project_id in self.running_processes:
            del self.running_processes[project_id]
            logger.info(f"Task for project {project_id} removed from pending area")
        else:
            logger.error(
                f"Failed to find task for project {project_id} in pending area"
            )

    def reset_pending_tasks(self):
        """Reset all pending tasks - terminates running subprocesses."""
        if self.running_processes:
            logger.info(f"Resetting {len(self.running_processes)} pending tasks")

            # Terminate all running processes
            for project_id, process in self.running_processes.items():
                try:
                    if process.is_alive():
                        logger.info(f"Terminating process for project {project_id}")
                        process.terminate()
                        # Give process 5 seconds to terminate gracefully, then kill
                        process.join(timeout=5.0)
                        if process.is_alive():
                            logger.warning(
                                f"Force killing process for project {project_id}"
                            )
                            process.kill()
                            process.join()
                except Exception as e:
                    logger.error(
                        f"Failed to terminate process for project {project_id}: {e}"
                    )

            self.running_processes.clear()
            logger.info("All pending tasks terminated and cleared")
        else:
            logger.info("No pending tasks to reset")

    def move_from_waiting_to_pending(self, project_id, process):
        record = self.is_waiting(project_id)
        if record:
            try:
                # add to pending
                self.add_pending(project_id, process)
                # delete
                self.session.execute(text("BEGIN TRANSACTION"))
                self.session.delete(record)
                self.session.commit()
                logger.info(f"Task for project {project_id} moved to pending area")
            except Exception:
                self.session.rollback()
                # remove from pending
                self.remove_pending(project_id)
                logger.error(
                    f"Failed to move task for project {project_id} to pending area"
                )

    def add_pending(self, project_id, process):
        if project_id not in self.running_processes:
            self.running_processes[project_id] = process

    def __execute_job(self, project_id, simulation):
        try:
            # run the simulation / train task
            p = RunModelProcess(
                func=run_task,
                args=(project_id, simulation),
                host=self.host,
                port=self.port,
            )
            p.start()
            logger.info(f"Task for project {project_id} started in a subprocess")
            return p  # Return the process object
        except Exception as _:
            message = f"Failed to spin up training process for project: {project_id}"
            logger.error(message)
            return None

    def _count_available_slots(self):
        return self.max_workers - len(self.running_processes)

    def pop_waiting_queue(self):
        """Moves tasks from the database and executes them in subprocess."""
        # Do we have available slots?
        if self._count_available_slots() > 0:
            # select first #max_workers records to ensure we
            # have new projects that can start training (note that
            # we can have only one record per project in the waiting
            # table, and that the same project may exist in pending).
            # If I have 3 slots available out of 8, selecting 8 records
            # ensures I will have 3 non-pending projects in the selection.
            try:
                records = (
                    self.session.query(ProjectQueueModel)
                    .order_by(ProjectQueueModel.id)
                    .limit(self.max_workers)
                    .all()
                )
            except Exception as e:
                logger.error(f"Failed to select waiting project ids: {e}")
                # just wait until the next tick
                records = []

            # loop over records
            for record in records:
                project_id = record.project_id
                simulation = record.simulation

                if project_id in self.running_processes:
                    # continue if this project is already in pending
                    continue
                elif self._count_available_slots() == 0:
                    # break if we have no more slots
                    break
                else:
                    # we have a slot and a new project,
                    # execute job:
                    process = self.__execute_job(project_id, simulation)
                    if process:
                        # move out of waiting and put into pending
                        self.move_from_waiting_to_pending(project_id, process)

    def _process_buffer(self):
        """Injects messages in the database."""
        while self.message_buffer:
            message = self.message_buffer.popleft()
            action = message.get("action", False)
            project_id = message.get("project_id", False)
            simulation = message.get("simulation", False)

            if action == "insert" and project_id:
                # This will insert into the waiting database if
                # the project isn't there, it will fail gracefully
                # if the project is already waiting
                self.insert_in_waiting(project_id, simulation)

            elif action in ["remove", "failure"] and project_id:
                if action == "failure":
                    logger.error(f"Failed to train model for project {project_id}")
                self.remove_pending(project_id)

            elif action == "reset_pending":
                self.reset_pending_tasks()

    def _handle_incoming_messages(self, conn):
        """Handles incoming traffic."""
        client_buffer = ""
        while True:
            try:
                data = conn.recv(self.receive_bytes)
                logger.debug(f"{data}")
                if not data:
                    # Process any remaining buffered messages on disconnect
                    if client_buffer.strip():
                        client_buffer = self._process_complete_messages(
                            client_buffer, conn
                        )
                    # client disconnected
                    break

                else:
                    message = data.decode("utf-8")
                    client_buffer += message

                    # Try to process complete JSON messages immediately
                    client_buffer = self._process_complete_messages(client_buffer, conn)

            except Exception as e:
                logger.info(f"Error while receiving message:\n{e}\n")
                break

        conn.close()

    def _process_complete_messages(self, client_buffer, conn):
        """Process complete JSON messages immediately, return remaining buffer."""
        # do not try to parse empty message or if we know in advance we did not
        # receive the entire message
        stripped_buffer = client_buffer.strip()
        if not (stripped_buffer and stripped_buffer.endswith("}")):
            return client_buffer

        try:
            # Try to parse as complete JSON message(s)
            temp_buffer = "[" + client_buffer.replace("}{", "},{") + "]"
            messages = json.loads(temp_buffer)

            # Process status queries immediately, queue others
            for message in messages:
                if message.get("action") == "status_query":
                    logger.debug(f"Processing immediate status query: {message}")
                    self._send_status_response(conn)
                else:
                    # Queue non-status messages for later processing
                    self.message_buffer.append(message)

            # All complete messages have been processed, clear the buffer
            return ""

        except json.JSONDecodeError:
            # JSON is incomplete, keep buffering
            return client_buffer

    def _send_status_response(self, conn):
        """Send status response back to client."""
        try:
            status = {
                "max_workers": self.max_workers,
                "currently_running": len(self.running_processes),
                "available_slots": self._count_available_slots(),
                "running_project_ids": list(self.running_processes.keys()),
            }
            response = json.dumps(status).encode("utf-8")
            conn.sendall(response)
            logger.debug(f"Sent status response: {status}")
        except Exception as e:
            logger.error(f"Failed to send status response: {e}")

    def _bind_server_socket(self, mp_start_event=None):
        """Bind the server socket to the configured host and port."""
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self.server_socket.bind((self.host, self.port))
        except OSError as e:
            if e.errno in (48, 98, 10048):
                if self.server_socket:
                    self.server_socket.close()
                    self.server_socket = None

                if mp_start_event is not None:
                    logger.warning(
                        "Address already in use, socket "
                        f"reused on {self.host}:{self.port}"
                    )
                    mp_start_event.set()
                    return False

            logger.exception(f"Failed to bind socket: {e}")
            return False

        logger.info(f"Socket bound to {self.host}:{self.port}")
        return True

    def start_manager(self, mp_start_event=None, mp_shutdown_event=None):
        """Start the task manager.

        Parameters
        ----------
        mp_start_event : multiprocessing.Event
            Event to signal that the manager has started.
        mp_shutdown_event : multiprocessing.Event
            Event to signal that the manager should shut down.

        """
        logger.info(f"Starting task server on {self.host}:{self.port}")

        if not self._bind_server_socket(mp_start_event):
            return

        self.server_socket.listen()

        if mp_start_event is not None:
            mp_start_event.set()

        self.server_socket.settimeout(1.0)

        def _signal_handler(signum, frame):
            logger.info(f"Shutting down task manager due to signal {signum}")
            self.stop_manager(mp_shutdown_event)

        signal.signal(signal.SIGINT, _signal_handler)
        signal.signal(signal.SIGTERM, _signal_handler)

        while mp_shutdown_event is None or not mp_shutdown_event.is_set():
            try:
                # Accept incoming connections with a timeout
                conn, _ = self.server_socket.accept()
                # Start a new thread to handle the client connection
                client_thread = threading.Thread(
                    target=self._handle_incoming_messages, args=(conn,)
                )
                # client_thread.daemon = True
                client_thread.start()
                # Store the latest client thread and conn
                self.client_thread = client_thread
                self.client_conn = conn

            except socket.timeout:
                # No incoming connections => perform handling queue
                self._process_buffer()
                # Pop tasks from database into 'pending'
                self.pop_waiting_queue()
                # continue to check for shutdown conditions
                continue

            except OSError as e:
                if e.errno == errno.EBADF:
                    # Socket was closed, exit loop silently
                    break
                logger.error(f"Socket error occurred: {e}")
                break  # Exit the loop if the socket is closed

        # After exiting main loop, clean up client thread
        if self.client_thread and self.client_conn:
            try:
                self.client_conn.shutdown(socket.SHUT_RDWR)
            except Exception:
                pass
            try:
                self.client_conn.close()
            except Exception:
                pass
            self.client_thread.join(timeout=1.0)

    def stop_manager(self, mp_shutdown_event=None):
        """Gracefully stop the manager and close the socket."""

        # Signal to the main process that the manager is shutting down
        # and set the shutdown event if provided
        if mp_shutdown_event is not None and not mp_shutdown_event.is_set():
            logger.info("Shutting down task manager...")
            mp_shutdown_event.set()

        # Close the server socket
        if self.server_socket:
            try:
                self.server_socket.close()
            except OSError as e:
                logger.error(f"Failed to close server socket: {e}")
            finally:
                self.server_socket = None
        logger.info("Task manager has been stopped.")

        # Close the database session

        if self.session:
            try:
                self.session.close()
            except Exception as e:
                logger.error(f"Failed to close database session: {e}")
            self.session = None

        # After closing server socket and DB session, also clean up client thread
        if getattr(self, "client_thread", None) and getattr(self, "client_conn", None):
            try:
                self.client_conn.shutdown(socket.SHUT_RDWR)
            except Exception:
                pass
            try:
                self.client_conn.close()
            except Exception:
                pass
            self.client_thread.join(timeout=1.0)


def run_task_manager(
    max_workers=None,
    host=None,
    port=None,
    mp_start_event=None,
    mp_shutdown_event=None,
    verbose=0,
):
    kwargs = {}
    if max_workers is not None:
        kwargs["max_workers"] = max_workers
    if host is not None:
        kwargs["host"] = host
    if port is not None:
        kwargs["port"] = port

    _setup_logging(verbose=verbose)

    manager = TaskManager(**kwargs)
    manager.start_manager(mp_start_event, mp_shutdown_event)



import base64
import hashlib
import hmac

from flask import Blueprint
from flask import current_app
from flask import jsonify
from flask import request
from flask_login import current_user
from sqlalchemy.exc import SQLAlchemyError

import asreview as asr
from asreview.webapp import DB
from asreview.webapp._authentication.decorators import login_required
from asreview.webapp._authentication.models import Project
from asreview.webapp._authentication.models import User
from asreview.webapp.utils import get_project_path

bp = Blueprint("team", __name__, url_prefix="/api")

REQUESTER_FRAUD = {"message": "Request can not made by current user."}


def validate_invitation_token(encoded_token):
    """
    Validate and decode an invitation token.

    Returns tuple of (project, error_response).
    If validation succeeds, returns (project, None).
    If validation fails, returns (None, (json_response, status_code)).
    """
    if not encoded_token:
        return None, (jsonify({"message": "No invitation token provided."}), 400)

    try:
        # Decode the base64-encoded token
        decoded = base64.urlsafe_b64decode(encoded_token.encode("utf-8"))

        # Split payload and signature based on position
        # SHA256 signature is always 32 bytes at the end, preceded by a 1-byte separator
        if len(decoded) < 34:  # At least 1 byte payload + separator + 32 byte signature
            return None, (
                jsonify({"message": "Token is not valid."}),
                400,
            )

        signature = decoded[-32:]
        separator = decoded[-33:-32]
        payload_bytes = decoded[:-33]

        # Verify separator is a dot
        if separator != b".":
            return None, (
                jsonify({"message": "Token is not valid."}),
                400,
            )

        # Verify HMAC signature
        secret_key = current_app.config.get("SECRET_KEY", "").encode("utf-8")
        expected_signature = hmac.new(
            secret_key, payload_bytes, hashlib.sha256
        ).digest()

        if not hmac.compare_digest(signature, expected_signature):
            return None, (
                jsonify({"message": "Token is not valid."}),
                400,
            )

        # Extract project_id and token from payload
        payload = payload_bytes.decode("utf-8")
        if ":" not in payload:
            return None, (
                jsonify({"message": "Token is not valid."}),
                400,
            )

        project_id, token = payload.split(":", 1)

        # Find the project
        project = Project.query.filter(Project.project_id == project_id).one_or_none()

        if not project:
            return None, (jsonify({"message": "Project not found."}), 404)

        # Verify the token matches the project's token
        if not project.token or project.token != token:
            return None, (
                jsonify({"message": "This invitation link is not valid."}),
                400,
            )

        return project, None

    except (ValueError, base64.binascii.Error):
        return None, (jsonify({"message": "Token is not valid."}), 400)
    except Exception as e:
        return None, (jsonify({"message": f"Error: {str(e)}"}), 500)


def get_user_project_properties(user, project, current_user):
    """Retrieve user properties in relation to a project and current user."""
    result = user.summarize()

    owner = user.id == project.owner_id
    member = user in project.collaborators or owner
    me = user.id == current_user.id

    deletable = (
        (current_user.id == project.owner_id or current_user.is_admin)
        and member
        and not me
    )

    return dict(
        result,
        me=me,
        owner=owner,
        member=member,
        deletable=deletable,
    )


@bp.route("/projects/<project_id>/users", methods=["GET"])
@login_required
def users(project_id):
    """Returns users involved in a project (owner and members)."""
    project = Project.query.filter(Project.project_id == project_id).one_or_none()

    # Deny if user is member and this person is not somehow involved in the project
    if (
        current_user.is_member
        and project not in current_user.projects
        and project not in current_user.involved_in
    ):
        return jsonify(REQUESTER_FRAUD), 404

    # Get only users involved in the project
    involved_users = set()

    # Add project owner
    if project.owner:
        involved_users.add(project.owner)

    # Add collaborators (members)
    involved_users.update(project.collaborators)

    users = sorted(
        [
            get_user_project_properties(user, project, current_user)
            for user in involved_users
        ],
        key=lambda u: (-u["owner"], u["name"].lower()),
    )

    return (
        jsonify(users),
        200,
    )


@bp.route("/projects/<project_id>/users/<user_id>", methods=["DELETE"])
@login_required
def end_collaboration(project_id, user_id):
    """Project owner or admin removes a collaborator, or collaborator
    removes him/herself."""
    response = jsonify(REQUESTER_FRAUD), 404
    # get project
    project = Project.query.filter(Project.project_id == project_id).one_or_none()

    # check if project is owned by current user, if the user is
    # involved in the project, or if current user is admin
    if project and (
        (project.owner == current_user)
        or (project in current_user.involved_in)
        or current_user.is_admin
    ):
        user = DB.session.get(User, user_id)

        if not user:
            return jsonify({"message": "User not found."}), 404

        # Prevent removing project owner
        if user.id == project.owner_id:
            return jsonify({"message": "Cannot remove project owner."}), 400

        try:
            if user in project.collaborators:
                project.collaborators.remove(user)
                DB.session.commit()
                response = (
                    jsonify(
                        {
                            "message": "Member removed successfully.",
                            "user": get_user_project_properties(
                                user, project, current_user
                            ),
                        }
                    ),
                    200,
                )
            else:
                response = (jsonify({"message": "User is not a collaborator."}), 400)

        except SQLAlchemyError as e:
            response = (
                jsonify({"message": f"Error removing collaborator: {str(e)}"}),
                500,
            )
    return response


@bp.route("/team/join/preview", methods=["POST"])
@login_required
def preview_invitation():
    """Get project details from invitation token without joining."""
    data = request.get_json()
    encoded_token = data.get("encoded_token")

    # Validate the invitation token
    project, error_response = validate_invitation_token(encoded_token)
    if error_response:
        return error_response

    # Get project name
    project_path = get_project_path(project.project_id)
    asr_project = asr.Project(project_path, project_id=project.project_id)
    project_name = asr_project.config.get("name", project.project_id)

    return (
        jsonify(
            {
                "project_id": project.project_id,
                "project_name": project_name,
                "owner_name": project.owner.name if project.owner else "Unknown",
                "is_owner": current_user.id == project.owner_id,
                "is_member": current_user in project.collaborators,
            }
        ),
        200,
    )


@bp.route("/team/join", methods=["POST"])
@login_required
def join_project():
    """Join a project using an invitation token."""
    data = request.get_json()
    encoded_token = data.get("encoded_token")

    # Validate the invitation token
    project, error_response = validate_invitation_token(encoded_token)
    if error_response:
        return error_response

    try:
        # Load ASReview project to get project name
        project_path = get_project_path(project.project_id)
        asr_project = asr.Project(project_path, project_id=project.project_id)
        project_name = asr_project.config.get("name", project.project_id)

        # Check if user is already the owner
        if current_user.id == project.owner_id:
            return (
                jsonify(
                    {
                        "message": "You are already the owner of this project.",
                        "already_member": True,
                        "project_id": project.project_id,
                        "project_name": project_name,
                    }
                ),
                200,
            )

        # Check if user is already a member
        if current_user in project.collaborators:
            return (
                jsonify(
                    {
                        "message": "You are already a member of this project.",
                        "already_member": True,
                        "project_id": project.project_id,
                        "project_name": project_name,
                    }
                ),
                200,
            )

        # Add user to collaborators
        project.collaborators.append(current_user)
        DB.session.commit()

        return (
            jsonify(
                {
                    "message": "Successfully joined the project.",
                    "project_id": project.project_id,
                    "project_name": project_name,
                }
            ),
            200,
        )

    except SQLAlchemyError as e:
        DB.session.rollback()
        return jsonify({"message": f"Error joining project: {str(e)}"}), 500



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils import check_random_state

from asreview.models.mixin import QueryMixin

__all__ = [
    "Max",
    "HybridMaxRandom",
    "HybridMaxUncertainty",
    "Uncertainty",
    "Random",
    "TopDown",
]


def _random_array(n, random_state=None):
    row_indices = np.arange(n)
    check_random_state(random_state).shuffle(row_indices)
    return row_indices


def _mix_indices(query_idx_1, query_idx_2, mix_probability=0.95, random_state=None):
    query_idx_mix = []
    i = 0
    j = 0

    while i < len(query_idx_1) and j < len(query_idx_2):
        if check_random_state(random_state).rand() < mix_probability:
            query_idx_mix.append(query_idx_1[i])
            i = i + 1
        else:
            query_idx_mix.append(query_idx_2[j])
            j = j + 1

    indexes = np.unique(query_idx_mix, return_index=True)[1]
    return [query_idx_mix[i] for i in sorted(indexes)]


class Random(QueryMixin, BaseEstimator):
    """Random query strategy.

    Choose the samples to be included at random.
    """

    name = "random"
    label = "Random"

    def __init__(self, random_state=None):
        self.random_state = random_state

    def query(self, p):
        """Query instances.

        Arguments
        ---------
        p: np.array
            The probabilities of the instances.
        random_state: int, RandomState
            Random state for shuffling the indices.

        Returns
        -------
        np.array:
            The indices of the instances to be queried.
        """
        return _random_array(len(p), random_state=self.random_state)


class TopDown(QueryMixin, BaseEstimator):
    """Top-down query strategy.

    Query the records in a top-down fashion.
    """

    name = "top_down"
    label = "Top-down"

    def query(self, p):
        """Query instances.

        Arguments
        ---------
        p: np.array
            The probabilities of the instances.

        Returns
        -------
        np.array:
            The indices of the instances to be queried.
        """
        return np.arange(len(p))


class Uncertainty(QueryMixin, BaseEstimator):
    """Uncertainty query strategy.

    Choose the most uncertain samples according to the model (i.e. closest to
    0.5 probability). If decision functions are used, the samples closest to
    the decision boundary are chosen (0).

    """

    name = "uncertainty"
    label = "Uncertainty"

    def __init__(self, u: float = None, proba: bool = True):
        self.u = u
        self.proba = proba

    def query(self, p):
        """Query instances.

        Arguments
        ---------
        p: np.array
            The probabilities of the instances.

        Returns
        -------
        np.array:
            The indices of the instances to be queried.
        """

        if self.u is None and self.proba:
            u = 0.5
        elif self.u is None and not self.proba:
            u = 0
        else:
            u = self.u

        try:
            return np.argsort(np.abs(p - u), kind="stable")
        except TypeError:
            raise TypeError("Probabilities or decision functions should be provided")


class Max(QueryMixin, BaseEstimator):
    """Maximum query strategy.

    Choose the most likely samples to be included according to the model.
    """

    name = "max"
    label = "Maximum"

    def query(self, p):
        try:
            return np.argsort(-p, kind="stable")
        except TypeError:
            raise TypeError("Probabilities or decision functions should be provided")


class HybridMaxUncertainty(QueryMixin, BaseEstimator):
    """95% Maximum and 5% Uncertainty query strategy.

    A mix of maximum and random query strategies with a mix ratio of 0.95.
    At each query 95% of the instances would be sampled with the maximum
    query strategy after which the remaining 5% would be sampled with
    the uncertainty query strategy.
    """

    name = "max_uncertainty"
    label = "Mixed (95% Maximum and 5% Uncertainty)"

    def __init__(self, probability=0.95, u=None, proba=None, random_state=None):
        self.probability = probability
        self.u = u
        self.proba = proba
        self.random_state = random_state

    def query(self, p):
        return _mix_indices(
            Max().query(p),
            Uncertainty(u=self.u, proba=self.proba).query(p),
            self.probability,
            self.random_state,
        )


class HybridMaxRandom(QueryMixin, BaseEstimator):
    """95% Maximum and 5% Random query strategy.

    A mix of maximum and random query strategies with a mix ratio of 0.95.
    At each query 95% of the instances would be sampled with the maximum
    query strategy after which the remaining 5% would be sampled with
    the random query strategy.

    Arguments
    ---------
    probability: float
        The probability of sampling with the maximum query strategy.
    random_state: int, RandomState
        Random

    """

    name = "max_random"
    label = "Mixed (95% Maximum and 5% Random)"

    def __init__(self, probability=0.95, random_state=None):
        self.probability = probability
        self.random_state = random_state

    def query(self, p):
        return _mix_indices(
            Max().query(p),
            _random_array(len(p), self.random_state),
            self.probability,
            self.random_state,
        )



from pathlib import Path
import shutil
import urllib.request

import pandas as pd
import pytest

from asreview import Project


_OSF_CACHE_DIR = Path("~/.cache/asreview_tests").expanduser()
_OSF_FG93A_URL = "https://osf.io/download/fg93a/"


@pytest.fixture(scope="session")
def osf_fg93a_path():
    """Return a local path to the OSF fg93a dataset, downloading it if needed."""
    _OSF_CACHE_DIR.mkdir(parents=True, exist_ok=True)

    # Return cached file without making any network requests
    cached_files = list(_OSF_CACHE_DIR.iterdir())
    if cached_files:
        return cached_files[0]

    # No cached file found, download it
    from asreview.utils import _get_filename_from_url

    filename = _get_filename_from_url(_OSF_FG93A_URL)
    cache_file = _OSF_CACHE_DIR / filename
    urllib.request.urlretrieve(_OSF_FG93A_URL, cache_file)
    return cache_file


@pytest.fixture
def demo_data(render_data=False):
    """Get a demo dataset.

    Returns
    -------
    pd.DataFrame:
        A demo dataset with 10 inclusions and 90 exclusions. Inclusions are found on
        every 10th row. The dataset is a sample from the van der Waal 2022 dataset.
    """

    if render_data:
        from synergy_dataset import Dataset

        df = Dataset("van_der_Waal_2022").to_frame(["title", "abstract", "open_access"])
        df_is_oa = df[df["open_access"].apply(lambda x: x["is_oa"])]

        df_inclusions = df_is_oa[df_is_oa["label_included"] == 1].sample(
            10, replace=False, random_state=165
        )

        df_is_oa = df_is_oa[df_is_oa["label_included"] == 0].sample(
            100, replace=False, random_state=165
        )
        df_is_oa.iloc[::-10] = df_inclusions

        df_is_oa.to_json(Path("tests", "demo_data.json"), orient="records", lines=True)
        # df_is_oa.to_csv(Path("tests", "demo_data.csv"), index=False)
        return df_is_oa

    return pd.read_json(Path("tests", "demo_data.json"), orient="records", lines=True)


@pytest.fixture
def demo_data_path(demo_data, tmp_path):
    """Get path to a demo dataset stored in a temporary file.

    Returns
    -------
    Path:
        Path to CSV file containing the demo dataset.
    """
    file_path = tmp_path / "demo_data.csv"
    demo_data.to_csv(file_path, index=False)
    return file_path


@pytest.fixture
def tmp_project(tmpdir):
    """Get a temporary project path.

    Returns
    -------
    Path:
        Path to a temporary project.
    """

    return Path(tmpdir, "test.asreview")


@pytest.fixture
def asreview_test_project_path(tmpdir):
    """Fixture to set up a test project for ASReview."""
    test_state_fp = Path("tests", "asreview_files", "asreview-demo-project.asreview")
    tmp_project_path = Path(tmpdir, "asreview-demo-project.asreview")
    shutil.copy(test_state_fp, tmp_project_path)
    return tmp_project_path


@pytest.fixture
def asreview_test_project(asreview_test_project_path, tmpdir):
    unzip_path = Path(tmpdir, "unzipped_project")
    return Project.load(asreview_test_project_path, unzip_path)



import json
import sqlite3
import time
from functools import cached_property

import pandas as pd

from asreview.data.record import Record
from asreview.database.store import DataStore
from asreview.database.store import _build_conn_uri

__all__ = ["Database"]

CURRENT_DATABASE_VERSION = 3

MODEL_COLUMNS = [
    "classifier",
    "querier",
    "balancer",
    "feature_extractor",
    "training_set",
]

REQUIRED_TABLES = [
    "results",
    "last_ranking",
    "decision_changes",
]

RESULTS_TABLE_COLUMNS_PANDAS_DTYPES = {
    "record_id": "Int64",
    "label": "Int64",
    "classifier": "object",
    "querier": "object",
    "balancer": "object",
    "feature_extractor": "object",
    "training_set": "Int64",
    "time": "Float64",
    "note": "object",
    "tags": "object",
    "user_id": "Int64",
}

RANKING_TABLE_COLUMNS_PANDAS_DTYPES = {
    "record_id": "Int64",
    "ranking": "Int64",
    "classifier": "object",
    "querier": "object",
    "balancer": "object",
    "feature_extractor": "object",
    "training_set": "Int64",
    "time": "Float64",
}


def open_db(fp, read_only=False):
    """Open a database.

    Parameters
    ----------
    fp : path-like
        File path to the database
    read_only : bool, optional
        Whether to create a new database if one doesn't exist yet and whether the opened
        database will be in read only mode or not.

    Returns
    -------
    Database
        ASReview database.

    Raises
    ------
    FileNotFoundError
        If `read_only` and there is no file at `fp`.
    ValueError
        If `read_only` and there is no valid database at `fp`.
    """
    if not fp.is_file():
        if read_only:
            raise FileNotFoundError(
                f"File path {fp} is not a file and 'read_only' is 'True'"
            )
        fp.parent.mkdir(parents=True, exist_ok=True)

    db = Database(fp, read_only=read_only)
    try:
        db._is_valid()
    except ValueError as e:
        if read_only:
            raise ValueError(
                f"There is no valid database at {fp} and the database is opened in"
                " read-only mode"
            ) from e
        db.create_tables()
    return db


class Database:
    """Database containing the input data and results.

    Database contains two parts: the input and the results. For more information on the
    input, see `asreview.database.store.py`. For more information on the results, see
    `asreview.database.sqlstate.py`.

    Attributes
    ----------
    user_version: str
        Return the version number of the database.
    """

    def __init__(self, fp=":memory:", record_cls=Record, read_only=False):
        """Initialize the Database.

        Parameters
        ----------
        fp : str | Path
            Path of the database file. Use `":memory:"` for an in-memory database.
        record_cls : type[asreview.data.record.Base], optional
            Type to use for the input records, see `DataStore` for more information.
        read_only : bool, optional
            Whether to open the database in read only mode. If the database is opened in
            read only mode and an attempt to write to the database is made, an
            `sqlite3.OperationalError` will be raised.
        """
        if fp == ":memory:" and read_only:
            raise ValueError("Can't open an in-memory database in read only mode")

        self.fp = fp
        self.record_cls = record_cls
        self.read_only = read_only
        self._in_memory = fp == ":memory:"
        self._closed = False
        self._conn_uri = _build_conn_uri(fp, read_only)

        self.input = DataStore(
            conn_uri=self._conn_uri, record_cls=record_cls, read_only=read_only
        )

        if self._in_memory:
            # Eagerly open the sqlite3 connection. For named in-memory databases,
            # the database is destroyed when the last connection to it closes.
            # This connection acts as an anchor that keeps the database alive
            # for the lifetime of this object.
            self._conn

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    @cached_property
    def _conn(self):
        """Get a connection to the SQLite database.

        Returns
        -------
        sqlite3.Connection
            Connection to the SQLite database.
        """
        return sqlite3.connect(self._conn_uri, uri=True)

    def close(self):
        """Close the database and release all resources.

        For in-memory databases this will destroy the database. Safe to call multiple
        times.
        """
        if self._closed:
            return
        self._closed = True
        self.input.engine.dispose()
        if "_conn" in self.__dict__:
            self._conn.close()
            del self.__dict__["_conn"]

    @property
    def user_version(self):
        """Version number of the state."""
        cur = self._conn.cursor()
        version = cur.execute("PRAGMA user_version")

        return int(version.fetchone()[0])

    @user_version.setter
    def user_version(self, version):
        cur = self._conn.cursor()
        cur.execute(f"PRAGMA user_version = {version}")
        self._conn.commit()
        cur.close()

    def create_tables(self):
        self.user_version = CURRENT_DATABASE_VERSION
        self.input.create_tables()

        cur = self._conn.cursor()

        cur.execute(
            """CREATE TABLE results
                            (record_id INTEGER UNIQUE,
                            label INTEGER,
                            classifier TEXT,
                            querier TEXT,
                            balancer TEXT,
                            feature_extractor TEXT,
                            training_set INTEGER,
                            time FLOAT,
                            note TEXT,
                            tags JSON,
                            user_id INTEGER)"""
        )

        cur.execute(
            """CREATE TABLE last_ranking
                            (record_id INTEGER UNIQUE,
                            ranking INT,
                            classifier TEXT,
                            querier TEXT,
                            balancer TEXT,
                            feature_extractor TEXT,
                            training_set INTEGER,
                            time FLOAT)"""
        )

        cur.execute(
            """CREATE TABLE decision_changes
                            (record_id INTEGER,
                            label INTEGER,
                            time FLOAT,
                            user_id INTEGER)"""
        )

        self._conn.commit()

        self._set_results_changes_triggers()

    def _is_valid(self):
        if self.user_version != CURRENT_DATABASE_VERSION:
            raise ValueError(
                f"Database version {self.user_version} is not supported. "
                "See migration guide."
            )
        cur = self._conn.cursor()
        column_names = cur.execute("PRAGMA table_info(results)").fetchall()
        table_names = cur.execute(
            "SELECT name FROM sqlite_master WHERE type='table';"
        ).fetchall()

        table_names = [tup[0] for tup in table_names]
        missing_tables = [
            table
            for table in REQUIRED_TABLES + [self.record_table_name]
            if table not in table_names
        ]
        if missing_tables:
            raise ValueError(
                f"The SQL file should contain tables named "
                f"'{' '.join(missing_tables)}'."
            )

        column_names = [tup[1] for tup in column_names]
        missing_columns = [
            col
            for col in RESULTS_TABLE_COLUMNS_PANDAS_DTYPES.keys()
            if col not in column_names
        ]
        if missing_columns:
            raise ValueError(
                f"The results table does not contain the columns "
                f"{' '.join(missing_columns)}."
            )

    def _set_results_changes_triggers(self):
        con = self._conn
        cur = con.cursor()
        cur.execute("""
            CREATE TRIGGER IF NOT EXISTS trg_results_delete
            AFTER DELETE ON results
            FOR EACH ROW
            BEGIN
                INSERT INTO decision_changes (record_id, label, time, user_id)
                VALUES (OLD.record_id, OLD.label, OLD.time, OLD.user_id);
            END
        """)
        cur.execute("""
            CREATE TRIGGER IF NOT EXISTS trg_results_label_update
            AFTER UPDATE OF label ON results
            FOR EACH ROW
            WHEN OLD.label IS NOT NEW.label
            BEGIN
                INSERT INTO decision_changes (record_id, label, time, user_id)
                VALUES (OLD.record_id, OLD.label, OLD.time, OLD.user_id);
            END
        """)
        con.commit()

    @property
    def record_table_name(self):
        return self.input.record_cls.__tablename__

    @property
    def exist_new_labeled_records(self):
        """Return True if there are new labeled records.

        Return True if there are any record labels added since the last time
        the model ranking was added to the state. Also returns True if no
        model was trained yet, but priors have been added.
        """
        labeled = self.get_results_table("label")
        last_training_set = self.get_last_ranking_table()["training_set"]

        if last_training_set.empty or pd.isna(last_training_set.max()):
            return len(labeled) > 0
        else:
            return len(labeled) > last_training_set.max()

    def _replace_results_from_df(self, results):
        if not set(results.columns) == set(RESULTS_TABLE_COLUMNS_PANDAS_DTYPES):
            raise ValueError(
                f"Columns of the results dataframe should be "
                f"{list(RESULTS_TABLE_COLUMNS_PANDAS_DTYPES.keys())}."
            )

        cur = self._conn.cursor()
        cur.execute("delete from results")
        self._conn.commit()
        cur.close()

        results.to_sql("results", self._conn, if_exists="append", index=False)

    def _replace_last_ranking_from_df(self, last_ranking):
        if not set(last_ranking.columns) == set(RANKING_TABLE_COLUMNS_PANDAS_DTYPES):
            raise ValueError(
                f"Columns of the last ranking dataframe should be "
                f"{list(RANKING_TABLE_COLUMNS_PANDAS_DTYPES.keys())}."
            )

        last_ranking.to_sql(
            "last_ranking", self._conn, if_exists="replace", index=False
        )

    def add_last_ranking(
        self,
        ranked_record_ids,
        classifier,
        querier,
        balancer,
        feature_extractor,
        training_set=None,
    ):
        """Save the ranking of the last iteration of the model.

        Save the ranking of the last iteration of the model, in the ranking
        order, so the record on row 0 is ranked first by the model.

        Parameters
        ----------
        ranked_record_ids: list, numpy.ndarray
            A list of records ids in the order that they were ranked.
        classifier: str
            Name of the classifier of the model.
        querier: str
            Name of the query strategy of the model.
        balancer: str
            Name of the balance strategy of the model.
        feature_extractor: str
            Name of the feature extraction method of the model.
        training_set: int
            Number of labeled records available at the time of training.
        """

        pd.DataFrame(
            {
                "record_id": ranked_record_ids,
                "ranking": range(len(ranked_record_ids)),
                "classifier": classifier,
                "querier": querier,
                "balancer": balancer,
                "feature_extractor": feature_extractor,
                "training_set": training_set,
                "time": time.time(),
            }
        ).to_sql("last_ranking", self._conn, if_exists="replace", index=False)

    def get_last_ranking_table(self):
        """Get the ranking from the state.

        Returns
        -------
        pd.DataFrame
            Dataframe with columns 'record_id', 'ranking', 'classifier',
            'querier', 'balancer', 'feature_extractor',
            'training_set' and 'time'. It has one row for each record in the
            dataset, and is ordered by ranking.
        """
        return pd.read_sql_query(
            "SELECT * FROM last_ranking",
            self._conn,
            dtype=RANKING_TABLE_COLUMNS_PANDAS_DTYPES,
        )

    def label_record(self, record_id, label, tags=None, user_id=None):
        if tags is not None:
            tags = json.dumps(tags)
        labeling_time = time.time()
        con = self._conn
        cur = con.cursor()
        model_string = ", ".join(MODEL_COLUMNS)
        target_result_string = ", ".join(
            f"target_result.{col}" for col in MODEL_COLUMNS
        )
        upsert_columns = ["label", "time", "tags", "user_id"] + MODEL_COLUMNS
        upsert_string = ", ".join(f"{col} = excluded.{col}" for col in upsert_columns)

        cur.execute(
            f"""
            WITH target_group AS (
                SELECT record_id
                FROM {self.record_table_name}
                WHERE group_id = (
                    SELECT group_id
                    FROM {self.record_table_name}
                    WHERE record_id=:record_id
                )
            ), target_result AS (
                SELECT {model_string}
                FROM results
                WHERE record_id = :record_id
            )
            INSERT INTO results(record_id, label, time, tags, user_id, {model_string})
            SELECT target_group.record_id, :label, :time, :tags, :user_id, {target_result_string}
            FROM target_group
            LEFT JOIN target_result ON 1
            ON CONFLICT(record_id) DO UPDATE
                SET {upsert_string};
            """,
            {
                "record_id": record_id,
                "label": label,
                "time": labeling_time,
                "tags": tags,
                "user_id": user_id,
            },
        )
        con.commit()

    def query_top_ranked(self, user_id=None):
        model_string = ", ".join(MODEL_COLUMNS)
        top_record_string = ", ".join(f"top_record.{col}" for col in MODEL_COLUMNS)
        con = self._conn
        cur = con.cursor()
        cur.execute(
            f"""INSERT INTO results (record_id, user_id, {model_string})
            WITH top_record AS (
                SELECT last_ranking.*
                FROM last_ranking
                LEFT JOIN results USING (record_id)
                WHERE results.record_id IS NULL
                ORDER BY ranking
                LIMIT 1
            ), group_records AS (
                SELECT record.record_id
                FROM record
                WHERE group_id = (
                    SELECT group_id
                    FROM record
                    WHERE record.record_id = (SELECT record_id FROM top_record)
                )
            )
            SELECT group_records.record_id, :user_id, {top_record_string}
            FROM group_records
            CROSS JOIN top_record;""",
            {"user_id": user_id},
        )
        con.commit()
        if not cur.rowcount:
            raise ValueError("Failed to query top ranked record")
        return self.get_pending(user_id=user_id)

    def update_result(self, record_id, label=None, tags=None, user_id=None):
        if label is None and tags is None:
            raise ValueError("At least one of 'label' or 'tags' must be provided.")

        fields = []
        values = {"record_id": record_id}
        if label is not None:
            fields.append("label = :label")
            values["label"] = label
            if user_id is not None:
                # We only update the user_id if the label changes.
                fields.append("user_id = :user_id")
                values["user_id"] = user_id
        if tags is not None:
            fields.append("tags = :tags")
            values["tags"] = json.dumps(tags)
        set_string = ", ".join(fields)

        con = self._conn
        cur = con.cursor()
        cur.execute(
            f"""
            WITH target_group AS (
                SELECT record_id
                FROM {self.record_table_name}
                WHERE group_id = (
                    SELECT group_id
                    FROM {self.record_table_name}
                    WHERE record_id=:record_id
                )
            )
            UPDATE results
            SET {set_string}
            WHERE record_id IN (SELECT record_id FROM target_group)
            """,
            values,
        )
        con.commit()

    def update_note(self, record_id, note=None):
        """Change the note of an already labeled or pending record.

        Parameters
        ----------
        record_id: int
            Id of the record whose label should be changed.
        note: str
            Note to add to the record.
        """

        cur = self._conn.cursor()
        cur.execute(
            f"""
            WITH target_group AS (
                SELECT record_id
                FROM {self.record_table_name}
                WHERE group_id = (
                    SELECT group_id
                    FROM {self.record_table_name}
                    WHERE record_id=:record_id
                )
            )
            UPDATE results SET note = :note WHERE record_id IN (
                SELECT record_id FROM target_group
            )""",
            {"note": note, "record_id": record_id},
        )

        if cur.rowcount == 0:
            raise ValueError(f"Record with id {record_id} not found.")

        self._conn.commit()

    def delete_result(self, record_id):
        con = self._conn
        cur = con.cursor()
        cur.execute(
            f"""
            WITH target_group AS (
                SELECT record_id
                FROM {self.record_table_name}
                WHERE group_id = (
                    SELECT group_id
                    FROM {self.record_table_name}
                    WHERE record_id=:record_id
                )
            )
            DELETE FROM results
            WHERE record_id IN (SELECT record_id FROM target_group)
            """,
            {"record_id": record_id},
        )
        con.commit()

    def get_results_record(self, record_id):
        """Get the data of a specific query from the results table.

        Parameters
        ----------
        record_id: int
            Record id of which you want the data.

        Returns
        -------
        pd.DataFrame
            Dataframe containing the data from the results table with the given
            record_id and columns.
        """

        result = pd.read_sql_query(
            f"SELECT * FROM results WHERE record_id={record_id}",
            self._conn,
            dtype=RESULTS_TABLE_COLUMNS_PANDAS_DTYPES,
        )
        result["tags"] = result["tags"].map(json.loads, na_action="ignore")
        return result

    def get_results_table(self, columns=None, priors=True, pending=False, groups=False):
        """Get a subset from the results table.

        Can be used to get any column subset from the results table.
        Most other get functions use this one, except some that use a direct
        SQL query for efficiency.

        Parameters
        ----------
        columns: list, str
            List of columns names of the results table, or a string containing
            one column name.
        priors: bool
            Whether to keep the records containing the prior knowledge.
        pending: bool
            Whether to keep the records which are pending a labeling decision.
        groups: bool
            Return all the records of a group of records. Be default only returns the
            base record of each group.

        Returns
        -------
        pd.DataFrame:
            Dataframe containing the data of the specified columns of the
            results table.
        """
        if isinstance(columns, str):
            columns = [columns]

        if (not priors) or (not pending) or (not groups):
            sql_where = []
            if not priors:
                sql_where.append("querier is not NULL")
            if not pending:
                sql_where.append("label is not NULL")
            if not groups:
                sql_where.append(
                    f"record_id IN ( SELECT group_id FROM {self.record_table_name})"
                )
            sql_where_str = "WHERE " + " AND ".join(sql_where)
        else:
            sql_where_str = ""

        if columns is None:
            col_dtype = RESULTS_TABLE_COLUMNS_PANDAS_DTYPES
        else:
            col_dtype = {
                k: v
                for k, v in RESULTS_TABLE_COLUMNS_PANDAS_DTYPES.items()
                if columns and k in columns
            }

        query_string = "*" if columns is None else ",".join(columns)
        df_results = pd.read_sql_query(
            f"SELECT {query_string} FROM results {sql_where_str} ORDER BY rowid",
            self._conn,
            dtype=col_dtype,
        )

        if columns is None or "tags" in columns:
            df_results["tags"] = df_results["tags"].map(json.loads, na_action="ignore")
        return df_results

    def get_priors(self):
        """Get the record ids of the priors.

        Returns
        -------
        pd.DataFrame:
            The result records of the priors in the order they were added. If multiple
            records are in the same group, only the base record of the group is
            returned.
        """
        df_results = pd.read_sql_query(
            f"""
            SELECT * FROM results
            WHERE results.querier is NULL
            AND results.label is not NULL
            AND record_id IN (
                SELECT group_id FROM {self.record_table_name}
            )
            ORDER BY rowid
            """,
            self._conn,
            dtype=RESULTS_TABLE_COLUMNS_PANDAS_DTYPES,
        )
        df_results["tags"] = df_results["tags"].map(json.loads, na_action="ignore")
        return df_results

    def get_pool(self):
        """Get the unlabeled, not-pending records in ranking order.

        Returns
        -------
        pd.Series
            Series containing the record_ids of the unlabeled, not pending
            records, in the order of the last available ranking. If the state does not
            yet contain a last ranking, the return value will be an empty dataframe. If
            multiple records are in the same group, only the base record of the group is
            returned.
        """

        return pd.read_sql_query(
            f"""SELECT record_id, last_ranking.ranking
                FROM last_ranking
                LEFT JOIN results
                USING (record_id)
                WHERE results.record_id is null AND last_ranking.record_id IN (
                    SELECT group_id FROM {self.record_table_name}
                )
                ORDER BY ranking
                """,
            self._conn,
        )["record_id"]

    def get_unlabeled(self, groups=False):
        """Get the unlabeled record ids in ranking order.

        Records that have no label or no entry in the results table are considered
        unlabeled.

        Parameters
        ----------
        groups : bool
            If True, return all records in each unlabeled group. If False,
            return only group representatives (record_id == group_id).

        Returns
        -------
        pd.Series
            Series of record_ids of unlabeled records ordered by ranking.
        """
        if groups:
            sql_group_filter = ""
        else:
            sql_group_filter = (
                f"AND record_id IN (SELECT group_id FROM {self.record_table_name})"
            )

        return pd.read_sql_query(
            f"""SELECT record_id, last_ranking.ranking
            FROM last_ranking
            JOIN {self.record_table_name} USING (record_id)
            LEFT JOIN results USING (record_id)
            WHERE (results.record_id IS NULL OR results.label IS NULL)
            {sql_group_filter}
            ORDER BY ranking
            """,
            self._conn,
        )["record_id"]

    def get_pending(self, user_id=None):
        """Get pending records from the results table.

        Parameters
        ----------
        user_id: int
            User id of the user who labeled the records.

        Returns
        -------
        pd.DataFrame
            DataFrame with pending results records.
        """
        query = f"""SELECT * FROM results WHERE label is null AND record_id IN (
            SELECT group_id FROM {self.record_table_name}
        )"""
        params = None
        if user_id is not None:
            query += " AND user_id=?"
            params = (user_id,)
        query += " ORDER BY rowid"

        return pd.read_sql_query(
            query,
            self._conn,
            params=params,
            dtype=RESULTS_TABLE_COLUMNS_PANDAS_DTYPES,
        )

    def get_decision_changes(self):
        """Get the record ids for any decision changes.

        Get the record ids of the records whose labels have been changed after the
        original labeling action.

        Returns
        -------
        pd.DataFrame
            Dataframe with columns 'record_id', 'label', 'time', and 'user_id' for each
            record of which the labeling decision was changed.
        """

        return pd.read_sql_query("SELECT * FROM decision_changes", self._conn)



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["RISReader", "RISWriter"]

import copy
import io
import json
import re
from urllib.request import urlopen

import pandas as pd
import rispy

from asreview.data.base import BaseReader
from asreview.data.utils import convert_value_to_list
from asreview.utils import _is_url

RIS_NOTE_LABEL_MAPPING = {
    "ASReview_relevant": 1,
    "ASReview_irrelevant": 0,
    "ASReview_not_seen": None,
}
LABEL_RIS_NOTE_MAPPING = {val: key for key, val in RIS_NOTE_LABEL_MAPPING.items()}


def _parse_label_from_notes(note_list):
    if not isinstance(note_list, list):
        return
    for note in note_list:
        for key, val in RIS_NOTE_LABEL_MAPPING.items():
            if key in note:
                return val


def _remove_asreview_data_from_notes(note_list):
    """Remove ASReview data from notes.

    Parameters
    ----------
    note_list: list
        A list of notes, coming from the Dataframe's "notes" column.

    Returns
    -------
    asreview_new_notes: list
        A list of updated notes, where ASReview data has been removed.
    """

    # Return {} and an empty list
    if not isinstance(note_list, list):
        return []
    return [
        note
        for note in note_list
        if note not in RIS_NOTE_LABEL_MAPPING and not note.startswith("asreview_")
    ]


class RISReader(BaseReader):
    """RIS file reader."""

    read_format = [".ris", ".txt"]
    write_format = [".csv", ".tsv", ".xlsx", ".ris"]

    mime_types = {
        "application/x-research-info-systems": [".ris", ".txt"],
        "text/plain": [".txt", ".ris"],
    }

    def _strip_zotero_p_tags(note_list):
        """Converter function for removing the XHTML <p></p> tags from Zotero export.

        Parameters
        ----------
        note_list: list
            A list of notes, coming from the Dataframe's "notes" column.

        Returns
        -------
        new_notes: list
            A list of updated notes, where XHTML <p></p> tags have been stripped.
        note_list: list
            The original note_list, when no XHTML <p></p> tags have been found.
        """
        if isinstance(note_list, list):
            new_notes = []
            for v in note_list:
                try:
                    new_notes.append(re.sub(r"^<p>|<\/p>$", "", v))
                except Exception:
                    new_notes.append(v)
            return new_notes
        else:
            return note_list

    @classmethod
    def _read_from_file(cls, fp, encoding="utf8"):
        with open(fp, encoding=encoding) as bibliography_file:
            return list(rispy.load(bibliography_file, skip_unknown_tags=True))

    @classmethod
    def _read_from_url(cls, fp, encoding="utf8"):
        url_input = urlopen(fp)

        bibliography_file = io.StringIO(url_input.read().decode(encoding))
        entries = list(rispy.load(bibliography_file, skip_unknown_tags=True))
        bibliography_file.close()

        return entries

    @classmethod
    def read_data(cls, fp):
        """Import dataset.

        Parameters
        ----------
        fp: str, pathlib.Path
            File path to the RIS file.

        Returns
        -------
        pd.DataFrame:
            Dataframe with entries. If the notes field contains a note with the text
            `ASReview_relevant`, `ASReview_irrelevant` or `ASReview_not_seen`, the
            data frame will have a column `included` with the value `1`, `0` or `None`.

        Raises
        ------
        ValueError
            File with unrecognized encoding is used as input.
        """
        encodings = ["utf-8", "utf-8-sig", "ISO-8859-1"]
        entries = None
        for encoding in encodings:
            try:
                if _is_url(fp):
                    entries = cls._read_from_url(fp, encoding=encoding)
                    break
                else:
                    entries = cls._read_from_file(fp, encoding=encoding)
                    break
            except UnicodeDecodeError:
                continue
            except Exception as e:
                raise ValueError(f"Error reading RIS file: {e}")

        if entries is None:
            raise ValueError("Cannot find proper encoding for data file")

        # Turn the entries dictionary into a Pandas dataframe
        df = pd.DataFrame(entries)

        # Check if "notes" column is present
        if "notes" in df:
            # Strip Zotero XHTML <p> tags on "notes"
            df["notes"] = df["notes"].apply(cls._strip_zotero_p_tags)
            # Parse the labels from the notes if present.
            labels = df["notes"].apply(_parse_label_from_notes)
            if not labels.isna().all():
                df["included"] = labels
            df["notes"] = df["notes"].apply(_remove_asreview_data_from_notes)

        if "included" in df:
            df["included"] = df["included"].astype("Int64")

        return df

    @classmethod
    def clean_data(cls, df):
        # We drop the 'label' column if it's available. For RIS files ASReview stores
        # and loads the labels from the notes field.
        df.drop("label", axis=1, inplace=True, errors="ignore")

        # We combine the values of 'authors' and 'first_authors' into one list of
        # authors. Internally we only use the authors when searching for a record, so
        # it does not matter too much if combining the two lists leads to the wrong
        # order of authors, or duplicate authors appearing in the list.
        if "authors" not in df:
            df["authors"] = None
        if "first_authors" not in df:
            df["first_authors"] = None
        df["authors"] = df["authors"].apply(convert_value_to_list)
        df["first_authors"] = df["first_authors"].apply(convert_value_to_list)
        df["authors"] = df["authors"] + df["first_authors"]

        return super().clean_data(df)


class RISWriter:
    """RIS file writer."""

    name = "ris"
    label = "RIS"
    caution = "Available only if you imported a RIS file when creating the project"
    write_format = ".ris"

    @classmethod
    def write_data(cls, df, fp):
        """Export dataset.

        Parameters
        ----------
        df: pd.Dataframe
            Dataframe of all available record data.
        fp: str, pathlib.Path
            File path to the RIS file, if exists.

        Returns
        -------
        RIS file
            Dataframe of all available record data. Any column from the data frame that
            starts with `asreview_` is added to the RIS file as note in the notes field
            of the form: `asreview_{column_name}: json.dumps({column_value})`. If the
            dataframe contains a column `asreview_label`, also a note is added with the
            value `ASReview_relevant`, `ASReview_irrelevant` or `ASReview_not_seen`
            corresponding to the value `1`, `0` or `None` in that column.
        """
        # Turn pandas DataFrame into records (list of dictionaries) for rispy
        records = copy.deepcopy(df.to_dict("records"))

        # Create an array for storing modified records
        records_new = []

        # Iterate over all available records
        for rec in records:

            def _isnull(v):
                if isinstance(v, list):
                    return v == []

                return pd.isnull(v)

            rec_copy = {}
            rec_copy["notes"] = rec.pop("notes", [])
            for key, val in rec.items():
                if key == "asreview_label":
                    rec_copy["notes"].insert(
                        0, LABEL_RIS_NOTE_MAPPING[rec["asreview_label"]]
                    )
                elif _isnull(val):
                    continue
                elif key.startswith("asreview_"):
                    rec_copy["notes"].append(f"{key}: {json.dumps(val)}")
                else:
                    rec_copy[key] = val
            if rec_copy["notes"] == []:
                rec_copy.pop("notes")

            # Throw away columns that can not be exported to RIS.
            rec_copy = {
                key: val
                for key, val in rec_copy.items()
                if not key == "included" and not key.startswith("asreview_")
            }
            # Append the deepcopied and updated record to a new array
            records_new.append(rec_copy)

        # From buffered dataframe
        if fp is None:
            # Write the whole content to buffer
            return rispy.dumps(records_new)

        # From IO dataframe
        else:
            # Write the whole content to a file
            with open(fp, "w", encoding="utf8") as fp:
                rispy.dump(records_new, fp)



# Copyright 2019-2025 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# the schema describes the content of the ASReview project file.

SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "$id": "http://example.com/example.json",
    "type": "object",
    "title": "The ASReview project file root schema",
    "description": "The root schema comprises the entire project.json file in the ASReview project file.",
    "default": {},
    "examples": [
        {
            "version": "1.0",
            "project_file_version": 3,
            "id": "example",
            "mode": "oracle",
            "name": "example",
            "description": "",
            "authors": "",
            "created_at_unix": 1648205610,
            "review": {
                "status": "review",
                "model": {"name": "example_model", "current_value": {}},
            },
            "feature_matrices": [
                {"id": "tfidf", "filename": "tfidf_feature_matrix.npz"}
            ],
            "dataset_path": "example.ris",
            "tags": [
                {
                    "label": "Foo",
                    "export": "foo",
                    "values": [
                        {"label": "Bar", "export": "bar", "id": 0},
                    ],
                    "id": 0,
                }
            ],
        }
    ],
    "required": ["version", "project_file_version", "id", "mode", "name"],
    "properties": {
        "version": {
            "$id": "#/properties/version",
            "type": "string",
            "title": "The version schema",
            "description": "The version of ASReview on initiation of the project.",
            "default": "",
            "examples": ["1.0"],
        },
        "project_file_version": {
            "$id": "#/properties/project_file_version",
            "type": "integer",
            "title": "The project file version schema",
            "description": "The version number of the ASReview project file format.",
            "default": 1,
            "examples": [1],
        },
        "id": {
            "$id": "#/properties/id",
            "type": "string",
            "title": "The id schema",
            "description": "The unique identifier of the project.",
            "default": "",
            "examples": ["example"],
        },
        "mode": {
            "$id": "#/properties/mode",
            "type": "string",
            "title": "The mode schema",
            "description": "The mode of the project. One of oracle or simulate.",
            "default": "oracle",
            "enum": ["oracle", "simulate"],
            "examples": ["oracle"],
        },
        "name": {
            "$id": "#/properties/name",
            "type": ["string", "null"],
            "title": "The name schema",
            "description": "The name of the project.",
            "default": "",
            "examples": ["example"],
        },
        "description": {
            "$id": "#/properties/description",
            "type": ["string", "null"],
            "title": "The description schema",
            "description": "The description of the project.",
            "default": "",
            "examples": [""],
        },
        "authors": {
            "$id": "#/properties/authors",
            "type": ["string", "null"],
            "title": "The authors schema",
            "description": "The authors of the project.",
            "default": "",
            "examples": [""],
        },
        "created_at_unix": {
            "$id": "#/properties/created_at_unix",
            "type": ["integer", "null"],
            "title": "The created_at_unix schema",
            "description": "An explanation about the purpose of this instance.",
            "default": 0,
            "examples": [1648205610],
        },
        "review": {
            "$id": "#/properties/review",
            "type": "object",
            "title": "The review schema",
            "description": "The review in the project. Only a single review is allowed.",
            "default": {},
            "examples": [
                {
                    "status": "review",
                    "model": {"name": "example_model", "current_value": {}},
                }
            ],
            "required": ["status", "model"],
            "properties": {
                "status": {
                    "$id": "#/properties/review/properties/status",
                    "type": "string",
                    "title": "The status of the review.",
                    "description": "The status of the review. Options are setup, review, finished.",
                    "enum": ["setup", "review", "finished"],
                    "default": "setup",
                    "examples": ["review"],
                },
                "model": {
                    "$id": "#/properties/review/properties/model",
                    "type": "object",
                    "title": "The model schema",
                    "description": "The model configuration for the review.",
                    "default": {},
                    "properties": {
                        "name": {
                            "$id": "#/properties/review/properties/model/properties/name",
                            "type": "string",
                            "title": "The model name",
                            "description": "The name of the model.",
                            "examples": ["example_model"],
                        },
                        "current_value": {
                            "$id": "#/properties/review/properties/model/properties/current_value",
                            "type": "object",
                            "title": "The current value schema",
                            "description": "The current value of the model.",
                            "default": {},
                        },
                    },
                    "additionalProperties": False,
                },
            },
            "additionalProperties": False,
        },
        "feature_matrices": {
            "$id": "#/properties/feature_matrices",
            "type": "array",
            "title": "The feature_matrices schema",
            "description": "Information about the feature matrices.",
            "default": [],
            "examples": [[{"id": "tfidf", "filename": "tfidf_feature_matrix.npz"}]],
            "additionalItems": True,
            "items": {
                "$id": "#/properties/feature_matrices/items",
                "anyOf": [
                    {
                        "$id": "#/properties/feature_matrices/items/anyOf/0",
                        "type": "object",
                        "title": "The first anyOf schema",
                        "description": "Information about a feature matrix.",
                        "default": {},
                        "examples": [
                            {"id": "tfidf", "filename": "tfidf_feature_matrix.npz"}
                        ],
                        "required": ["id", "filename"],
                        "properties": {
                            "id": {
                                "$id": "#/properties/feature_matrices/items/anyOf/0/properties/id",
                                "type": "string",
                                "title": "The id schema",
                                "description": "A unique id of the feature matrix.",
                                "default": "",
                                "examples": ["tfidf"],
                            },
                            "filename": {
                                "$id": "#/properties/feature_matrices/items/anyOf/0/properties/filename",
                                "type": "string",
                                "title": "The filename schema",
                                "description": "The name of the file with the feature matrix. Usually a sparse matrix.",
                                "default": "",
                                "examples": ["tfidf_feature_matrix.npz"],
                            },
                        },
                        "additionalProperties": True,
                    }
                ],
            },
        },
        "tags": {
            "$id": "#/properties/tags",
            "title": "The tags schema",
            "description": "Optional tags for the project with no type restrictions.",
        },
        "datasets": {
            "$id": "#/properties/datasets",
            "type": "array",
            "title": "The datasets schema",
            "description": "The list of datasets in the project.",
            "default": [],
            "additionalItems": False,
            "items": {
                "$id": "#/properties/datasets/items",
                "anyOf": [
                    {
                        "$id": "#/properties/datasets/items/anyOf/0",
                        "type": "object",
                        "title": "The first anyOf schema",
                        "description": "An explanation about the purpose of this instance.",
                        "required": ["id", "name"],
                        "properties": {
                            "id": {
                                "$id": "#/properties/datasets/items/anyOf/0/properties/id",
                                "type": "string",
                                "title": "The id of the dataset.",
                                "description": "A unique UUID4 identifier of the dataset.",
                                "default": "",
                                "examples": ["4793de70a8d44eb4baa68bac2853c91a"],
                            },
                            "name": {
                                "$id": "#/properties/datasets/items/anyOf/0/properties/name",
                                "type": "string",
                                "title": "The filename of the dataset.",
                                "description": "The name of the dataset file.",
                                "default": "",
                                "examples": ["example.ris"],
                            },
                        },
                        "additionalProperties": False,
                        "examples": [
                            {
                                "id": "4793de70a8d44eb4baa68bac2853c91a",
                                "name": "example.ris",
                            }
                        ],
                    }
                ],
            },
            "examples": [
                [
                    {
                        "id": "4793de70a8d44eb4baa68bac2853c91a",
                        "name": "example.ris",
                    }
                ]
            ],
        },
    },
    "additionalProperties": True,
}
